# SPDX-License-Identifier: BSD-3-Clause
# Copyright (c) 2022 Scipp contributors (https://github.com/scipp)
# @author Simon Heybrock

# flake8: noqa
r"""
Specifics for time-of-flight neutron-scattering data reduction, including coordinate transformations.
"""

from .diagram import TimeDistanceDiagram
